# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-04-27 18:19
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('activity', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='activitygroup',
            name='context_object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='context_object_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='activity',
            name='status',
            field=models.CharField(choices=[('ACTIVE', 'ACTIVE'), ('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('SUCCESS', 'SUCCESS')], db_index=True, default='PENDING', max_length=255),
        ),
        migrations.AlterField(
            model_name='activitygroup',
            name='status',
            field=models.CharField(choices=[('ACTIVE', 'ACTIVE'), ('FAILURE', 'FAILURE'), ('SUCCESS', 'SUCCESS')], db_index=True, default='ACTIVE', max_length=255),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='steps_complete',
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='steps_total',
            field=models.IntegerField(null=True),
        ),
    ]
